@@ -1,3 +1,29 @@
+#' Create a server
+#'
+#' Create a server with a custom handler to handle the HTTP request.
+#' @param ... Arguments to be passed to \code{\link{server_config}()}.
+#' @param handler A function that takes the HTTP request and returns a response.
+#' @param ws_open A function to be called back when a WebSocket connection is
+#'   established (see \code{httpuv::\link{startServer}()}).
+#' @export
+#' @examplesIf interactive()
+#' # always return "Success:" followed by the requested path
+#' s = servr::create_server(handler = function(req) {
+#'   list(status = 200L, body = paste('Success:', req$PATH_INFO))
+#' })
+#' s$url
+#'
+#' browseURL(paste0(s$url, '/hello'))
+#' browseURL(paste0(s$url, '/world'))
+#'
+#' s$stop_server()
+create_server = function(..., handler, ws_open = function(ws) NULL) {
+  res = server_config(...)
+  app = list(call = handler, onWSOpen = ws_open)
+  res$start_server(app)
+  invisible(res)
+}
+
 #' Serve static files under a directory
 #'
 #' If there is an \file{index.html} under this directory, it will be displayed;
@@ -13,18 +39,14 @@
 #' @param ... server configurations passed to \code{\link{server_config}()}
 #' @export
 #' @references \url{https://github.com/yihui/servr}
-#' @examples #' see https://github.com/yihui/servr for command line usage
-#' # or run inside an R session
-#' if (interactive()) servr::httd()
+#' @examplesIf interactive()
+#' servr::httd()
 httd = function(dir = '.', ...) {
   dir = normalizePath(dir, mustWork = TRUE)
   if (dir != '.') {
     owd = setwd(dir); on.exit(setwd(owd))
   }
-  res = server_config(dir, ...)
-  app = list(call = serve_dir(dir))
-  res$start_server(app)
-  invisible(res)
+  create_server(dir, ..., handler = serve_dir(dir))
 }
 
 #' @param watch a directory under which \code{httw()} is to watch for changes;
@@ -119,15 +141,25 @@ watch_dir = function(
 #'   string to be used in the URL, e.g., \code{function(host) { if (host ==
 #'   '127.0.0.1') 'localhost' else host}} to convert \code{127.0.0.1} to
 #'   \code{localhost} in the URL.
+#' @param auth A list of the form \code{list(scheme, creds)} containing the
+#'   authentication scheme and credentials. See
+#'   \url{https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication} for
+#'   more info. Please note that this argument is \emph{by no means} intended
+#'   for serious HTTP applications and there is \emph{no warranty} on security.
+#'   You should use other dedicated software packages or services if security is
+#'   important. You have been warned.
 #' @param verbose Whether to print messages when launching the server.
 #' @inheritParams httpuv::startServer
 #' @export
 #' @return A list of configuration information of the form \code{list(host,
 #'   port, start_server = function(app) {}, ...)}.
+#' @examplesIf interactive()
+#' # an example of authentication
+#' servr::httd(auth = list(scheme = 'Basic', creds = servr::auth_basic('john', 'pa$s!')))
 server_config = function(
   dir = '.', host = getOption('servr.host', '127.0.0.1'), port, browser, daemon,
-  interval = getOption('servr.interval', 1), baseurl = '',
-  initpath = '', hosturl = identity, verbose = TRUE
+  interval = getOption('servr.interval', 1), baseurl = '', initpath = '',
+  hosturl = identity, auth = getOption('servr.auth'), verbose = TRUE
 ) {
   cargs = commandArgs(TRUE)
   if (missing(browser)) browser = interactive() || '-b' %in% cargs || is_rstudio()
@@ -142,8 +174,8 @@ server_config = function(
   # rstudio viewer cannot display a page served at 0.0.0.0; use 127.0.0.1 instead
   host2 = if (host == '0.0.0.0' && is_rstudio()) '127.0.0.1' else host
   url = sprintf('http://%s:%d', hosturl(host2), port)
-  baseurl = gsub('^/+', '', baseurl)
-  if (baseurl != '') url = paste0(url, '/', baseurl)
+  baseurl = gsub('^/*', '/', baseurl)
+  if (baseurl != '/') url = paste0(url, baseurl)
   url = paste0(url, if (initpath != '' && !grepl('^/', initpath)) '/', initpath)
   browsed = FALSE
   servrEnv$browse = browse = function(reopen = FALSE) {
@@ -153,9 +185,22 @@ server_config = function(
     if (verbose && !reopen) message('Serving the directory ', dir, ' at ', url)
   }
   server = NULL
+  # realm is required for Basic auth (append one in case it was not provided)
+  if (identical(tolower(auth$scheme), 'basic'))
+    auth$scheme = 'Basic realm="Access restricted"'
   list(
     host = host, port = port, interval = interval, url = url, daemon = daemon,
     start_server = function(app) {
+      if (is.function(app_call <- app$call)) {
+        app$call = function(req) {
+          # authentication
+          if (!auth_verify(req, auth)) return(list(
+            status = 401L, body = '', headers = list(`WWW-Authenticate` = auth$scheme)
+          ))
+          req = modify_path(req, baseurl)
+          app_call(req)
+        }
+      }
       id = startServer(host, port, app)
       if (verbose && daemon) daemon_hint(id); browse()
       server <<- id
@@ -172,6 +217,17 @@ server_config = function(
   )
 }
 
+# modify PATH_INFO in the request when baseurl is provided (remove baseurl)
+modify_path = function(req, baseurl) {
+  if (baseurl == '/') return(req)
+  path = decode_path(req)
+  if (startsWith(path, baseurl)) {
+    path = substr(path, nchar(baseurl) + 1, nchar(path))
+    req$PATH_INFO = httpuv::encodeURIComponent(path)
+  }
+  req
+}
+
 serve_dir = function(dir = '.') function(req) {
   owd = setwd(dir); on.exit(setwd(owd), add = TRUE)
   path = decode_path(req)
